/*
	*
	* Copyright 2019 Britanicus <marcusbritanicus@gmail.com>
	*
	* This file is a part of DesQ project (https://gitlab.com/desq/)
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>

#include <QDir>
#include <QProcess>
#include <QtDBus>

#include "DesQSession.hpp"
#include "DesQSessionAdaptor.hpp"

#include <desq/DesQUtils.hpp>

DesQSession::DesQSession( DesQSession::SessionType mType ) : QThread() {

	new SessionAdaptor( this );

	QDBusConnection dbusSess = QDBusConnection::sessionBus();
    registered &= dbusSess.registerObject( "/org/DesQ/Session", this );
    registered &= dbusSess.registerService( "org.DesQ.Session" );

	mSessionType = mType;

	switch( mSessionType ) {
		case DesQSession::DESQ: {
			session = new QSettings( "DesQ", "Session" );
			mSessionName = "DesQ";
			break;
		}

		case DesQSession::OYSTER: {
			session = new QSettings( "Oyster", "Session" );
			mSessionName = "Oyster";
			break;
		}
	}

	/* Power Interface */
	pwr = new SessionAdaptor( this );
};

bool DesQSession::isRunning() {

	return not registered;
};

void DesQSession::start( QString name ) {

	mSessionName = name;
	QThread::start();
};

void DesQSession::stop() {

	/* Current Session's ID */
	int XDG_SESSION_ID = qgetenv( "XDG_SESSION_ID" ).toInt();
	QString XDG_SESSION_ID_STR = QString::number( XDG_SESSION_ID );

	/* Close all processes started in this session */
	// Scan all the PIDs in /proc/, read the /proc/<pid>/environ
	// If there is an entry for XDG_SESSION_ID which matches our
	// Session's ID, close it.
	QStringList procs;
	QDir proc( "/proc/" );
	for( QString pid: proc.entryList( QStringList(), QDir::Dirs | QDir::NoDotAndDotDot ) ) {

		/* Ignore Wayfire PID */
		if ( pid.toInt() == wfPID )
			continue;

		QFile env( "/proc/" + pid + "/environ" );
		if ( env.open( QFile::ReadOnly ) ) {
			QByteArrayList envVars = env.readAll().simplified().split( '\x00' );
			for( QByteArray raw_envVar: envVars ) {
				QString envVar = QString::fromUtf8( raw_envVar );
				if ( envVar.startsWith( "XDG_SESSION_ID" ) and envVar.endsWith( XDG_SESSION_ID_STR ) ) {
					procs << pid;
					kill( pid.toInt(), SIGTERM );
				}
			}
		}
	}

	/* Wait time (in seconds) for the processes to close. Default 10s */
	for ( int i = 0; i < session->value( "WaitTimeOut", 10 ).toInt(); i++ ) {
		sleep( 1 );
		qApp->processEvents();
	}

	for( QString pid: procs )
		kill( pid.toInt(), SIGKILL );

	/* Standard way to close wayfire */
	int ret = kill( wfPID, SIGINT );
	if ( ret == -1 ) {
		qDebug() << "Failed to kill wayfire. Trying again.";
		system( ( QString( "kill %1" ).arg( wfPID ) ).toUtf8().constData() );
	}

	/* Remove temporary folder */
	DesQUtils::removeDir( qgetenv( "__DESQ_WORK_DIR" ) );

	if ( isRunning() )
		terminate();

	qApp->quit();
};

void DesQSession::run() {

	/* Setup the environmental variables */
	setupEnvironmentVars();

	/* Start the daemons */
	startDaemons();

	/* Start the window manager and hence the shell */
	startDesQShell();
};

void DesQSession::handleMessages( const QString msg ) {

	if ( msg == "logout" ) {

		qDebug() << "Trying to logout from the active session.";
		stop();
	}

	else if ( msg == "shutdown" ) {
		stop();
		pwr->Shutdown();
	}

	else if ( msg == "reboot" ) {
		stop();
		pwr->Reboot();
	}

	else if ( msg == "suspend" ) {
		pwr->Suspend();
	}

	else if ( msg == "hibernate" ) {
		pwr->Hibernate();
	}

	else if ( msg.startsWith( "--session" ) ) {
		#if QT_VERSION >= 0x050E01
		QStringList args = msg.split( " ", Qt::SkipEmptyParts );
		#else
		QStringList args = msg.split( " ", QString::SkipEmptyParts );
		#endif

		if ( args.count() == 2 )
			start( args[ 1 ] );

		else if ( args.count() == 1 )
			start();
	}
};

void DesQSession::setupEnvironmentVars() {

	/* Wayland or X11 */
	qputenv( "XDG_SESSION_TYPE", QByteArrayLiteral( "wayland" ) );

	/* Platform for Qt5 */
	qputenv( "QT_QPA_PLATFORM", QByteArrayLiteral( "wayland" ) );

	/* Qt5 theme to be used */
	qputenv( "QT_QPA_PLATFORMTHEME", QByteArrayLiteral( "qt5ct" ) );

	/* Revise: Wayfire SSD with windecor does quite well */
	qputenv( "QT_WAYLAND_DISABLE_DECORATION", "1" );
	qputenv( "QT_WAYLAND_DISABLE_WINDOWDECORATION", "1" );

	/* For Mozilla firefox and thunderbird */
	qputenv( "MOZ_ENABLE_WAYLAND", "1" );

	/* Desktop session */
	switch( mSessionType ) {
		case SessionType::DESQ: {
			qputenv( "DESKTOP_SESSION", "/usr/share/wayland-sessions/desqsession" );

    		/* Desktop session name */
    		qputenv( "XDG_SESSION_DESKTOP", "DesQ" );
	        qputenv( "XDG_CURRENT_DESKTOP", "DesQ" );

			wfConfig = QDir::home().filePath( ".config/DesQ/wayfire.ini" );
			qputenv( "WAYFIRE_CONFIG_FILE", wfConfig.toUtf8().constData() );
			if ( not QFile::exists( wfConfig ) ) {
				bool ok = true;
				ok &= QDir::home().mkpath( ".config/DesQ/" );
				ok &= QFile::copy( "/usr/share/desq/wayfire.ini", wfConfig );

				/* If we're unable to copy the config file to user's config dir, use the default */
				if ( not ok ) {
					qDebug() << "Error copying wayfire.ini to ~/.config/DesQ/";
					wfConfig = "/usr/share/desq/wayfire.ini";
				}
			}

			break;
		}

		case SessionType::OYSTER: {
			qputenv( "DESKTOP_SESSION", "/usr/share/wayland-sessions/oystersession" );
			qputenv( "WAYFIRE_CONFIG_FILE", "" );

    		/* Desktop session name */
    		qputenv( "XDG_SESSION_DESKTOP", "Oyster" );
	        qputenv( "XDG_CURRENT_DESKTOP", "Oyster" );

			break;
		}
	}

	/* Cursor theme */
	qputenv( "XCURSOR_THEME", session->value( "CursorTheme", "breeze_cursors" ).toByteArray() );
	qputenv( "XCURSOR_SIZE", session->value( "CursorSize", "24" ).toByteArray() );

	/* Temp workspace for DesQ */
	srand( time( NULL ) );

	/* Create the folder: /tmp/$USER */
	QString tmpPath = QString( "/tmp/" ) + qgetenv( "USER" ) + "/";
	mkdir( tmpPath.toLocal8Bit().constData(), 0700 );

	while ( true ) {
		int num = rand();
		if ( DesQUtils::exists( tmpPath + QString::number( num ) ) )
			continue;

		tmpPath += QString::number( num );
		break;
	}

	qDebug() << "Preparing session work directory" << tmpPath;
	mkdir( tmpPath.toUtf8().constData(), 0700 );
	qputenv( "__DESQ_WORK_DIR", tmpPath.toUtf8() );

	qputenv( "__DESQ_SESSION_NAME", mSessionName.toUtf8() );

	/* Custom Environment Variables - Should we load from .bashrc or .config/fish/config? */
	/* For now we do everything from our settings file, section [EnvVars]. Format: Var = value */
	session->beginGroup( "EnvVars" );
	for( QString var: session->allKeys() )
		qputenv( var.toUtf8(), session->value( var ).toString().toUtf8() );

	session->endGroup();
};

void DesQSession::startDaemons() {

	/* Start the DesQ Settings Daemon: This serves settings to the Shell and others */
	QProcess::startDetached( DesQUtils::getUtilityPath( "settings-daemon" ), {} );

	/* Start the DesQ Power Daemon: This takes actions based on the power status */
	QProcess::startDetached( DesQUtils::getUtilityPath( "power-daemon" ), {} );
};

void DesQSession::startDesQShell() {

	/* For DesQ */
	if ( mSessionType == DesQSession::DESQ ) {
		qDebug() << "Starting window manager: wayfire";

		/* Start wayfire. */
		qDebug() << "Starting wayfire" << QProcess::startDetached( "wayfire", { "--config", wfConfig }, QDir::homePath(), &wfPID );

		/* If wayfire isn't running, no point in continuing */
		if ( wfPID == 0 ) {
			qDebug() << "Wayfire is not running... Quitting the session";
			qApp->quit();
		}
	}
};
